"""
Configuration components for Skyrelis.

This module contains internal configuration classes and utilities.
Users typically won't import from this module directly.
"""
from .observer_config import ObserverConfig

__all__ = ["ObserverConfig"] 